Symmetric Key File Encryption
=============================

At a glance:

* **Cipher and Mode**: `AES-256-CTR`
* **Padding**: None (CTR mode doesn't pad)
* **Authentication**: `HMAC-SHA-256`
* **Construction**: `Encrypt then MAC`
* **Algorithm Backend**: `ext/openssl`

## `\Defuse\Crypto\File`

**WARNING:** This encryption library is not a silver bullet. It only provides
symmetric encryption given a uniformly random key. This means you **MUST NOT**
use an ASCII string like a password as the key parameter, it **MUST** be
a uniformly random key generated by `createNewRandomKey()`. If you want to
encrypt something with a password, apply a password key derivation function
like PBKDF2 or scrypt with a random salt to generate a key.

***WARNING:*** Error handling is *very* important, especially for crypto code! 

How to use this code:
=====================

Generating a Key
----------------

```php
try {
    $key = \Defuse\Crypto\File::createNewRandomKey();
    // WARNING: Do NOT encode $key with bin2hex() or base64_encode(),
    // they may leak the key to the attacker through side channels.
    // 
    // This library offers these two for cache-timing-safe encoding:
    //
    //     \Defuse\Crypto\File\binToHex()
    //     \Defuse\Crypto\File\hexToBin()
    //
} catch (\Defuse\Crypto\Exception\CryptoTestFailedException $ex) {
    die('Cannot safely create a key');
} catch (\Defuse\Crypto\Exception\CannotPerformOperationException $ex) {
    die('Cannot safely create a key');
}
```

Encrypting a File
-----------------

```php
/* 
  // Don't forget to generate a random key
  $key = \Defuse\Crypto\File::createNewRandomKey();
*/

$inputFilename = 'image.jpg';
$outputFilename = 'image.enc.jpg';

try {
    \Defuse\Crypto\File::encryptFile(
        $inputFilename,
        $outputFilename,
        $key
    );
} catch (\Defuse\Crypto\Exception\CryptoTestFailedException $ex) {
    die('Cannot safely perform encryption');
} catch (\Defuse\Crypto\Exception\CannotPerformOperationException $ex) {
    die('Cannot safely perform encryption');
}
```

Decrypting a File
-----------------

```php
/* 
  // Don't forget to generate a random key
  $key = \Defuse\Crypto\File::createNewRandomKey();
*/

$inputFilename = 'image.enc.jpg';
$outputFilename = 'image.dec.jpg';

try {
    \Defuse\Crypto\File::decryptFile(
        $inputFilename,
        $outputFilename,
        $key
    );
} catch (\Defuse\Crypto\Exception\CryptoTestFailedException $ex) {
    die('Cannot safely perform decryption');
} catch (\Defuse\Crypto\Exception\CannotPerformOperationException $ex) {
    die('Cannot safely perform decryption');
}
```

Encrypting a File Resource
--------------------------

```php
/* 
  // Don't forget to generate a random key
  $key = \Defuse\Crypto\File::createNewRandomKey();
*/

$iFile = \fopen('image2.jpg', 'rb');
$oFile = \fopen('image2.enc.jpg', 'wb');

try {
    \Defuse\Crypto\File::encryptResource($iFile, $oFile, $key);
} catch (\Defuse\Crypto\Exception\CryptoTestFailedException $ex) {
    die('Cannot safely perform encryption');
} catch (\Defuse\Crypto\Exception\CannotPerformOperationException $ex) {
    die('Cannot safely perform encryption');
}
```

Decrypting a File Resource
--------------------------

```php
/* 
  // Don't forget to generate a random key
  $key = \Defuse\Crypto\File::createNewRandomKey();
*/

$iFile = \fopen('image2.enc.jpg', 'rb');
$oFile = \fopen('image2.dec.jpg', 'wb');

try {
    \Defuse\Crypto\File::decryptResource($iFile, $oFile, $key);
} catch (\Defuse\Crypto\Exception\CryptoTestFailedException $ex) {
    die('Cannot safely perform decryption');
} catch (\Defuse\Crypto\Exception\CannotPerformOperationException $ex) {
    die('Cannot safely perform decryption');
}
```
